@@ -11,7 +11,7 @@
 %% The Original Code is RabbitMQ Federation.
 %%
 %% The Initial Developer of the Original Code is GoPivotal, Inc.
-%% Copyright (c) 2007-2013 GoPivotal, Inc.  All rights reserved.
+%% Copyright (c) 2007-2014 GoPivotal, Inc.  All rights reserved.
 %%
 
 -module(rabbit_federation_exchange_link).
@@ -29,10 +29,12 @@
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
          terminate/2, code_change/3]).
 
--import(rabbit_federation_util, [name/1, vhost/1]).
+-import(rabbit_misc, [pget/2]).
+-import(rabbit_federation_util, [name/1, vhost/1, pgname/1]).
 
 -record(state, {upstream,
                 upstream_params,
+                upstream_name,
                 connection,
                 channel,
                 consumer_tag,
@@ -129,6 +131,7 @@ handle_info({#'basic.deliver'{routing_key  = Key,
             State = #state{
               upstream            = Upstream = #upstream{max_hops = MaxH},
               upstream_params     = UParams,
+              upstream_name       = UName,
               downstream_exchange = #resource{name = XNameBin},
               downstream_channel  = DCh,
               channel             = Ch,
@@ -136,10 +139,13 @@ handle_info({#'basic.deliver'{routing_key  = Key,
     PublishMethod = #'basic.publish'{exchange    = XNameBin,
                                      routing_key = Key},
     %% TODO add user information here?
-    HeadersFun = fun (H) -> update_headers(UParams, Redelivered, H) end,
+    HeadersFun = fun (H) -> update_headers(UParams, UName, Redelivered, H) end,
     %% We need to check should_forward/2 here in case the upstream
     %% does not have federation and thus is using a fanout exchange.
-    ForwardFun = fun (H) -> rabbit_federation_util:should_forward(H, MaxH) end,
+    ForwardFun = fun (H) ->
+                         DName = rabbit_nodes:cluster_name(),
+                         rabbit_federation_util:should_forward(H, MaxH, DName)
+                 end,
     Unacked1 = rabbit_federation_link_util:forward(
                  Upstream, DeliverMethod, Ch, DCh, PublishMethod,
                  HeadersFun, ForwardFun, Msg, Unacked),
@@ -186,16 +192,16 @@ cast(Msg)        -> [gen_server2:cast(Pid, Msg) || Pid <- all()].
 cast(XName, Msg) -> [gen_server2:cast(Pid, Msg) || Pid <- x(XName)].
 
 join(Name) ->
-    pg2_fixed:create(Name),
-    ok = pg2_fixed:join(Name, self()).
+    pg2_fixed:create(pgname(Name)),
+    ok = pg2_fixed:join(pgname(Name), self()).
 
 all() ->
-    pg2_fixed:create(rabbit_federation_exchanges),
-    pg2_fixed:get_members(rabbit_federation_exchanges).
+    pg2_fixed:create(pgname(rabbit_federation_exchanges)),
+    pg2_fixed:get_members(pgname(rabbit_federation_exchanges)).
 
 x(XName) ->
-    pg2_fixed:create({rabbit_federation_exchange, XName}),
-    pg2_fixed:get_members({rabbit_federation_exchange, XName}).
+    pg2_fixed:create(pgname({rabbit_federation_exchange, XName})),
+    pg2_fixed:get_members(pgname({rabbit_federation_exchange, XName})).
 
 %%----------------------------------------------------------------------------
 
@@ -329,26 +335,31 @@ bind_cmd0(unbind, Source, Destination, RoutingKey, Arguments) ->
 %% restrictive max_hops we have yet passed through.
 
 update_binding(Args, #state{downstream_exchange = X,
-                            upstream            = Upstream}) ->
+                            upstream            = Upstream,
+                            upstream_name       = UName}) ->
     #upstream{max_hops = MaxHops} = Upstream,
     Hops = case rabbit_misc:table_lookup(Args, ?BINDING_HEADER) of
                undefined    -> MaxHops;
                {array, All} -> [{table, Prev} | _] = All,
                                {short, PrevHops} =
                                    rabbit_misc:table_lookup(Prev, <<"hops">>),
-                               lists:min([PrevHops - 1, MaxHops])
+                               case rabbit_federation_util:already_seen(
+                                      UName, All) of
+                                   true  -> 0;
+                                   false -> lists:min([PrevHops - 1, MaxHops])
+                               end
            end,
     case Hops of
         0 -> ignore;
-        _ -> Node = rabbit_federation_util:local_nodename(vhost(X)),
+        _ -> Cluster = rabbit_nodes:cluster_name(),
              ABSuffix = rabbit_federation_db:get_active_suffix(
                           X, Upstream, <<"A">>),
              DVHost = vhost(X),
              DName = name(X),
-             Down = <<Node/binary, ":", DVHost/binary,":", DName/binary, " ",
-                      ABSuffix/binary>>,
-             Info = [{<<"downstream">>, longstr, Down},
-                     {<<"hops">>,       short,   Hops}],
+             Down = <<DVHost/binary,":", DName/binary, " ", ABSuffix/binary>>,
+             Info = [{<<"cluster-name">>, longstr, Cluster},
+                     {<<"exchange">>,     longstr, Down},
+                     {<<"hops">>,         short,   Hops}],
              rabbit_basic:prepend_table_header(?BINDING_HEADER, Info, Args)
     end.
 
@@ -358,6 +369,13 @@ go(S0 = {not_started, {Upstream, UParams, DownXName}}) ->
     Unacked = rabbit_federation_link_util:unacked_new(),
     rabbit_federation_link_util:start_conn_ch(
       fun (Conn, Ch, DConn, DCh) ->
+              Props = pget(server_properties,
+                           amqp_connection:info(Conn, [server_properties])),
+              UName = case rabbit_misc:table_lookup(
+                             Props, <<"cluster_name">>) of
+                          {longstr, N} -> N;
+                          _            -> unknown
+                      end,
               {Serial, Bindings} =
                   rabbit_misc:execute_mnesia_transaction(
                     fun () ->
@@ -376,6 +394,7 @@ go(S0 = {not_started, {Upstream, UParams, DownXName}}) ->
                         consume_from_upstream_queue(
                           #state{upstream              = Upstream,
                                  upstream_params       = UParams,
+                                 upstream_name         = UName,
                                  connection            = Conn,
                                  channel               = Ch,
                                  next_serial           = Serial,
@@ -387,8 +406,6 @@ go(S0 = {not_started, {Upstream, UParams, DownXName}}) ->
               {noreply, State}
       end, Upstream, UParams, DownXName, S0).
 
-%% local / disconnected never gets invoked, see handle_info({'DOWN', ...
-
 consume_from_upstream_queue(
   State = #state{upstream            = Upstream,
                  upstream_params     = UParams,
@@ -480,9 +497,11 @@ ensure_internal_exchange(IntXNameBin,
                                internal    = true,
                                auto_delete = true},
     Purpose = [{<<"x-internal-purpose">>, longstr, <<"federation">>}],
+    XFUArgs = [{?MAX_HOPS_ARG,  long,    MaxHops},
+               {?NODE_NAME_ARG, longstr, rabbit_nodes:cluster_name()}
+               | Purpose],
     XFU = Base#'exchange.declare'{type      = <<"x-federation-upstream">>,
-                                  arguments = [{?MAX_HOPS_ARG, long, MaxHops} |
-                                               Purpose]},
+                                  arguments = XFUArgs},
     Fan = Base#'exchange.declare'{type      = <<"fanout">>,
                                   arguments = Purpose},
     rabbit_federation_link_util:disposable_connection_call(
@@ -492,7 +511,7 @@ ensure_internal_exchange(IntXNameBin,
 
 upstream_queue_name(XNameBin, VHost, #resource{name         = DownXNameBin,
                                                virtual_host = DownVHost}) ->
-    Node = rabbit_federation_util:local_nodename(DownVHost),
+    Node = rabbit_nodes:cluster_name(),
     DownPart = case DownVHost of
                    VHost -> case DownXNameBin of
                                 XNameBin -> <<"">>;
@@ -511,7 +530,11 @@ delete_upstream_exchange(Conn, XNameBin) ->
     rabbit_federation_link_util:disposable_channel_call(
       Conn, #'exchange.delete'{exchange = XNameBin}).
 
-update_headers(#upstream_params{table = Table}, Redelivered, Headers) ->
+update_headers(#upstream_params{table = Table}, UName, Redelivered, Headers) ->
     rabbit_basic:prepend_table_header(
-      ?ROUTING_HEADER, Table ++ [{<<"redelivered">>, bool, Redelivered}],
+      ?ROUTING_HEADER, Table ++ [{<<"redelivered">>, bool, Redelivered}] ++
+          header_for_name(UName),
       Headers).
+
+header_for_name(unknown) -> [];
+header_for_name(Name)    -> [{<<"cluster-name">>, longstr, Name}].
