@@ -44,14 +44,8 @@ process_frame(#mqtt_frame{ fixed = #mqtt_frame_fixed{ type = Type }},
   when Type =/= ?CONNECT ->
     {error, connect_expected, PState};
 process_frame(Frame = #mqtt_frame{ fixed = #mqtt_frame_fixed{ type = Type }},
-              PState ) ->
-    %%rabbit_log:info("MQTT received frame ~p ~n", [Frame]),
-    try process_request(Type, Frame, PState) of
-        Result -> Result
-    catch _:Error ->
-        close_connection(PState),
-        {error, Error}
-    end.
+              PState) ->
+    process_request(Type, Frame, PState).
 
 process_request(?CONNECT,
                 #mqtt_frame{ variable = #mqtt_frame_connect{
@@ -82,6 +76,7 @@ process_request(?CONNECT,
                             {?CONNACK_ACCEPT, Conn} ->
                                 link(Conn),
                                 {ok, Ch} = amqp_connection:open_channel(Conn),
+                                amqp_channel:enable_delivery_flow_control(Ch),
                                 ok = rabbit_mqtt_collector:register(
                                   ClientId, self()),
                                 Prefetch = rabbit_mqtt_util:env(prefetch),
@@ -211,10 +206,12 @@ amqp_callback({#'basic.deliver'{ consumer_tag = ConsumerTag,
                                  delivery_tag = DeliveryTag,
                                  routing_key  = RoutingKey },
                #amqp_msg{ props = #'P_basic'{ headers = Headers },
-                          payload = Payload }} = Delivery,
+                          payload = Payload },
+               DeliveryCtx} = Delivery,
               #proc_state{ channels      = {Channel, _},
                            awaiting_ack  = Awaiting,
                            message_id    = MsgId } = PState) ->
+    amqp_channel:notify_received(DeliveryCtx),
     case {delivery_dup(Delivery), delivery_qos(ConsumerTag, Headers, PState)} of
         {true, {?QOS_0, ?QOS_1}} ->
             amqp_channel:cast(
@@ -278,7 +275,8 @@ amqp_callback(#'basic.ack'{ multiple = false, delivery_tag = Tag },
     {ok, PState #proc_state{ unacked_pubs = gb_trees:delete(Tag, UnackedPubs) }}.
 
 delivery_dup({#'basic.deliver'{ redelivered = Redelivered },
-              #amqp_msg{ props = #'P_basic'{ headers = Headers }}}) ->
+              #amqp_msg{ props = #'P_basic'{ headers = Headers }},
+              _DeliveryCtx}) ->
     case rabbit_mqtt_util:table_lookup(Headers, <<"x-mqtt-dup">>) of
         undefined   -> Redelivered;
         {bool, Dup} -> Redelivered orelse Dup
@@ -489,7 +487,14 @@ amqp_pub(#mqtt_msg{ qos        = Qos,
 
 adapter_info(Sock, ProtoVer) ->
     amqp_connection:socket_adapter_info(
-             Sock, {'MQTT', integer_to_list(ProtoVer)}).
+             Sock, {'MQTT', human_readable_mqtt_version(ProtoVer)}).
+
+human_readable_mqtt_version(3) ->
+    "3.1.0";
+human_readable_mqtt_version(4) ->
+    "3.1.1";
+human_readable_mqtt_version(_) ->
+    "N/A".
 
 send_client(Frame, #proc_state{ socket = Sock }) ->
     %rabbit_log:info("MQTT sending frame ~p ~n", [Frame]),
