@@ -11,7 +11,7 @@
 %%  The Original Code is RabbitMQ.
 %%
 %%  The Initial Developer of the Original Code is GoPivotal, Inc.
-%%  Copyright (c) 2007-2013 GoPivotal, Inc.  All rights reserved.
+%%  Copyright (c) 2007-2014 GoPivotal, Inc.  All rights reserved.
 %%
 
 -module(rabbit_shovel_status).
@@ -19,7 +19,7 @@
 
 -export([start_link/0]).
 
--export([report/2, status/0]).
+-export([report/3, remove/1, status/0]).
 
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
          terminate/2, code_change/3]).
@@ -28,13 +28,16 @@
 -define(ETS_NAME, ?MODULE).
 
 -record(state, {}).
--record(entry, {name, info, timestamp}).
+-record(entry, {name, type, info, timestamp}).
 
 start_link() ->
     gen_server:start_link({local, ?SERVER}, ?MODULE, [], []).
 
-report(Name, Info) ->
-    gen_server:cast(?SERVER, {report, Name, Info, calendar:local_time()}).
+report(Name, Type, Info) ->
+    gen_server:cast(?SERVER, {report, Name, Type, Info, calendar:local_time()}).
+
+remove(Name) ->
+    gen_server:cast(?SERVER, {remove, Name}).
 
 status() ->
     gen_server:call(?SERVER, status, infinity).
@@ -46,12 +49,20 @@ init([]) ->
 
 handle_call(status, _From, State) ->
     Entries = ets:tab2list(?ETS_NAME),
-    {reply, [{Entry#entry.name, Entry#entry.info, Entry#entry.timestamp}
+    {reply, [{Entry#entry.name, Entry#entry.type, Entry#entry.info,
+              Entry#entry.timestamp}
              || Entry <- Entries], State}.
 
-handle_cast({report, Name, Info, Timestamp}, State) ->
-    true = ets:insert(?ETS_NAME, #entry{name = Name, info = Info,
+handle_cast({report, Name, Type, Info, Timestamp}, State) ->
+    true = ets:insert(?ETS_NAME, #entry{name = Name, type = Type, info = Info,
                                         timestamp = Timestamp}),
+    rabbit_event:notify(shovel_worker_status,
+                        split_name(Name) ++ split_status(Info)),
+    {noreply, State};
+
+handle_cast({remove, Name}, State) ->
+    true = ets:delete(?ETS_NAME, Name),
+    rabbit_event:notify(shovel_worker_removed, split_name(Name)),
     {noreply, State}.
 
 handle_info(_Info, State) ->
@@ -62,3 +73,12 @@ terminate(_Reason, _State) ->
 
 code_change(_OldVsn, State, _Extra) ->
     {ok, State}.
+
+split_status({running, MoreInfo})         -> [{status, running} | MoreInfo];
+split_status({terminated, Reason})        -> [{status, terminated},
+                                              {reason, Reason}];
+split_status(Status) when is_atom(Status) -> [{status, Status}].
+
+split_name({VHost, Name})           -> [{name,  Name},
+                                        {vhost, VHost}];
+split_name(Name) when is_atom(Name) -> [{name, Name}].
