@@ -125,7 +125,8 @@ unacknowledged_message_count() ->
     lists:sum([queue:len(C#cr.acktags) || C <- all_ch_record()]).
 
 add(ChPid, CTag, NoAck, LimiterPid, LimiterActive, Prefetch, Args, IsEmpty,
-    State = #state{consumers = Consumers}) ->
+    State = #state{consumers = Consumers,
+                   use       = CUInfo}) ->
     C = #cr{consumer_count = Count,
             limiter        = Limiter} = ch_record(ChPid, LimiterPid),
     Limiter1 = case LimiterActive of
@@ -144,7 +145,8 @@ add(ChPid, CTag, NoAck, LimiterPid, LimiterActive, Prefetch, Args, IsEmpty,
                          ack_required = not NoAck,
                          prefetch     = Prefetch,
                          args         = Args},
-    State#state{consumers = add_consumer({ChPid, Consumer}, Consumers)}.
+    State#state{consumers = add_consumer({ChPid, Consumer}, Consumers),
+                use       = update_use(CUInfo, active)}.
 
 remove(ChPid, CTag, State = #state{consumers = Consumers}) ->
     case lookup_ch(ChPid) of
